from ._kacherycloud_request import _kacherycloud_request


def _access_group_encrypt(text: str, *, access_group: str):
    payload = {
        'type': 'accessGroupEncrypt',
        'accessGroupId': access_group,
        'text': text
    }
    response = _kacherycloud_request(payload)
    return response['encryptedText']

def _access_group_decrypt(encrypted_text: str, *, access_group: str):
    payload = {
        'type': 'accessGroupDecrypt',
        'accessGroupId': access_group,
        'encryptedText': encrypted_text
    }
    response = _kacherycloud_request(payload)
    return response['decryptedText']


from typing import Any, Callable, Union
import requests
import simplejson

from ..get_project_id import get_project_id
from .._kacherycloud_request import _kacherycloud_request
from .upload_task_result import upload_task_result
from .._serialize import _serialize

def _run_task(*, task_type: str, task_name: str, task_job_id: str, task_function: Callable, task_input: Any, project_id: str, extra_kwargs: dict):
    _set_task_status(task_name=task_name, task_job_id=task_job_id, status='started', project_id=project_id)
    try:
        result = task_function(**task_input, **extra_kwargs)
        serialized_result = _serialize(result)
    except Exception as err:
        _set_task_status(task_name=task_name, task_job_id=task_job_id, status='error', error=str(err), project_id=project_id)
        return False
    if task_type == 'calculation':
        upload_task_result(task_name=task_name, task_job_id=task_job_id, serialized_task_result=serialized_result, project_id=project_id)
    _set_task_status(task_name=task_name, task_job_id=task_job_id, status='finished', project_id=project_id)
    return True

def _set_task_status(*, task_name: str, task_job_id: str, status: str, error: Union[str, None]=None, project_id: str):
    if status == 'error':
        print(f'Error in task {task_name}: {error}')
    print(f'Task {status}: {task_name}')
    message = {
        'type': 'setTaskStatus',
        'taskName': task_name,
        'taskJobId': task_job_id,
        'status': status
    }
    if error is not None:
        message['errorMessage'] = error
    else:
        if status == 'error':
            raise Exception('Status is error, but no error string provided.')
    if not project_id.startswith('local:'):
        payload = {
            'type': 'publishToPubsubChannel',
            'channelName': 'provideTasks',
            'message': message
        }
        
        if project_id is not None:
            payload['projectId'] = project_id
        else:
            payload['projectId'] = get_project_id()
        _kacherycloud_request(payload)
    else:
        port = _get_port_from_project_id(project_id)
        _publish_local_pubsub_message(channel=f'provideTasks', message=message, port=port)

def _get_port_from_project_id(project_id: str):
    if not project_id.startswith('local:'):
        raise Exception(f'Unexpected project_id in _get_port_from_project_id(): {project_id}')
    a = project_id.split(':')
    if len(a) != 2:
        raise Exception(f'Unexpected project_id in _get_port_from_project_id(): {project_id}')
    return int(a[1])

def _publish_local_pubsub_message(*, channel: str, message: dict, port: int):
    req = {
        'type': 'publish',
        'channel': channel,
        'messageBody': simplejson.dumps(message)
    }
    resp = requests.post(f'http://localhost:{port}', json=req)
    if resp.status_code != 200:
        raise Exception(f'Error publishing pubsub message')
    response = resp.json()
    assert response['type'] == 'publish', 'Unexpected response in publish pubsub message'


from typing import Union
from dataclasses import dataclass
import os
import requests
from .load_file import load_file_info, load_file_local
from ._kachery_gateway_request import _kachery_gateway_request

@dataclass
class RequestFileResult:
    found: Union[None, bool] = None
    queued: Union[None, bool] = None
    completed: Union[None, bool] = None
    running: Union[None, bool] = None
    local: bool = False
    errored: bool = False
    error_message: str = ''
    size: Union[None, int] = None
    bytes_uploaded: Union[None, int] = None

def request_file(uri: str, *, resource: str, timeout_sec: float, ignore_local: bool=False) -> RequestFileResult:
    if not ignore_local:
        a0 = load_file_local(uri)
        if a0 is not None:
            return RequestFileResult(
                found=True,
                size=os.path.getsize(a0),
                local=True
            )
    x0 = load_file_info(uri)
    if (x0 is not None) and (x0['found']):
        return RequestFileResult(
            found=True,
            size=x0['size'],
            completed=True
        )
    proxy_url = _get_proxy_url_for_resource(resource)
    req = {
        'type': 'requestFromClient',
        'resourceName': resource,
        'request': {
            'type': 'fileUpload',
            'uri': _remove_query_string_from_uri(uri)
        },
        'timeoutMsec': timeout_sec * 1000
    }
    resp = requests.post(proxy_url + '/api', json=req)
    if resp.status_code != 200:
        raise Exception(f'Error in requestFromClient: ({resp.status_code}) {resp.reason}: {resp.text}')
    response = resp.json()
    if response['type'] != 'responseToClient':
        raise Exception('Unexpected response to requestFromClient')
    if response.get('error', ''):
        raise Exception(f'Error requesting file: {response["error"]}')
    rr = response['response']
    if rr['type'] != 'fileUpload':
        raise Exception('Unexpected fileUpload response')
    status = rr['status']
    status_str = status['status'] # 'not-found' | 'queued' | 'running' | 'completed' | 'error'
    size = status.get('size', None)
    bytes_uploaded = status.get('bytesUploaded', None)
    error_str = status.get('error', '')
    return RequestFileResult(
        found = status_str in ['queued', 'running', 'completed'],
        queued = status_str == 'queued',
        completed = status_str == 'completed',
        running = status_str == 'running',
        errored = status_str == 'error',
        error_message = error_str,
        size = size,
        bytes_uploaded = bytes_uploaded
    )

_global = {
    'proxy_urls': {}
}

def _remove_query_string_from_uri(uri: str):
    i = uri.find('?')
    if i < 0: return uri
    return uri[:i]

def _get_proxy_url_for_resource(resource: str):
    if resource in _global['proxy_urls']:
        return _global['proxy_urls'][resource]
    kachery_zone = os.environ.get('KACHERY_ZONE', 'default')
    resp = _kachery_gateway_request({
        'type': 'getResourceInfo',
        'resourceName': resource,
        'zone': kachery_zone
    })
    found = resp['found']
    if not found:
        raise Exception(f'Resource not found in zone: {resource}')
    proxy_url = resp['resource']['proxyUrl']
    _global['proxy_urls'][resource] = proxy_url
    return proxy_url



from ._access_group_encrypt import _access_group_encrypt, _access_group_decrypt


def encrypt_uri(uri: str, *, access_group: str):
    if uri.startswith('sha1://'):
        a = uri.split('?')
        if len(a) < 2: a.append('')
        querystr = a[1]
        b = a[0].split('/')
        sha1 = b[2]
        sha1_enc = _access_group_encrypt(sha1, access_group=access_group)
        querystr_plus = f'?{querystr}' if len(querystr) > 0 else ''
        return f'sha1-enc://{sha1_enc}.ag_{access_group}{querystr_plus}'
    elif uri.startswith('sha1-enc://'):
        aa = uri.split('?')[0].split('/')[2]
        bb = aa.split('.')
        assert len(bb) == 2
        access_group_str = bb[1]
        assert access_group_str.startswith('ag_')
        access_group0 = access_group_str[3:]
        if access_group0 == access_group:
            return uri
        uri2 = decrypt_uri(uri)
        return encrypt_uri(uri2, access_group=access_group)
    else:
        raise Exception(f'Not able to encrypt URI: {uri}')

def decrypt_uri(uri: str):
    aa = uri.split('?')[0].split('/')[2]
    bb = aa.split('.')
    assert len(bb) == 2
    sha1_enc = bb[0]
    access_group_str = bb[1]
    assert access_group_str.startswith('ag_')
    access_group = access_group_str[3:]
    sha1 = _access_group_decrypt(sha1_enc, access_group=access_group)
    qstr = f'?{uri.split("?")[1]}' if len(uri.split('?')) > 1 else ''
    return f'sha1://{sha1}{qstr}'


import time
from typing import List
import simplejson
from .PubsubListener import MessageCallback
import requests


class LocalPubsubListener:
    def __init__(self, *, channel: str, port: int) -> None:
        self._channel = channel
        self._port = port
        self._message_callbacks: List[MessageCallback] = []
        self._stopped = False
        self._last_message_id = ''
    def on_message(self, callback: MessageCallback):
        self._message_callbacks.append(callback)
    def start(self):
        pass
    def wait(self, timeout_sec: float):
        req = {'type': 'subscribe', 'channel': self._channel, 'timeoutMsec': timeout_sec * 1000, 'lastMessageId': self._last_message_id}
        failed_at_least_once = False
        while True:
            try:
                resp = requests.post(f'http://localhost:{self._port}', json=req)
                if resp.status_code != 200:
                    raise Exception(f'Error in pubsub request')
                if failed_at_least_once:
                    print('Connected to local pubsub server')
                break
            except Exception as err:
                print('Error connecting to local pubsub server. Trying again in 5 seconds.', str(err))
                failed_at_least_once = True
                time.sleep(5)
        response = resp.json()
        messages = response['messages']
        for m in messages:
            message_body = m['messageBody']
            message_id = m['messageId']
            message_json = simplejson.loads(message_body)
            self._last_message_id = message_id
            for cb in self._message_callbacks:
                cb(channel=self._channel, message=message_json)
        if len(messages) == 0:
            time.sleep(timeout_sec)
    def stop(self):
        self._stopped = True
        time.sleep(1)


import os
import socket
import urllib.parse

from kachery_cloud.get_kachery_cloud_dir import get_kachery_cloud_dir

from .get_client_id import get_client_id
from ._client_keys import _sign_message_as_client
from ._get_kachery_gateway_url import _get_kachery_gateway_url
from ._kachery_gateway_request import _kachery_gateway_request
from ._get_local_client_config import _get_local_client_config
from ._kacherycloud_request import _kacherycloud_request


_global_init = {
    'client_info': 0 # 0 means we have not yet queried. None means not found.
}

def _get_client_info():
    if _global_init['client_info'] != 0:
        return _global_init['client_info']
    client_id = get_client_id()
    kachery_zone = os.environ.get('KACHERY_ZONE', 'default')
    payload = {
        'type': 'getClientInfo',
        'clientId': client_id,
        'zone': kachery_zone
    }
    response = _kachery_gateway_request(payload)
    # response = _kacherycloud_request(payload)
    found = response['found']
    if found:
        _global_init['client_info'] = response
        return response
    else:
        return None

def init():
    client_info = _get_client_info()
    if client_info is None:
        client_id = get_client_id()
        signature = _sign_message_as_client({
            'type': 'addClient'
        })
        label = socket.gethostname()
        kachery_zone = os.environ.get('KACHERY_ZONE', 'default')
        url = f'{_get_kachery_gateway_url()}/registerClient/{client_id}?signature={signature}&zone={kachery_zone}&label={urllib.parse.quote(label)}'
        print('')
        print(url)
        print('')
        print('Click the above link to register this kachery cloud client. After pressing the "REGISTER CLIENT" button on the website, press [Enter] in this terminal to continue.')
        print('')
        input()
        print('Checking...')
        client_info = _get_client_info()
        if client_info is None:
            raise Exception('Failed to initialize this kachery client')
        print(f'Client initialized successfully.')
    else:
        client_id = get_client_id()
        print('This client has already been registered.')
        print('Click the following link to configure the client:')
        url = f'{_get_kachery_gateway_url()}/client/{client_id}'
        print(url)
        print('')
    client = client_info['client']
    label = client['label']
    client_owner = client['ownerId']
    print(f'Client ID: {client_id}')
    print(f'Label: {label}')
    print(f'Owner: {client_owner}')

    print('')
    print('* Kachery-cloud is intended for collaborative sharing of data for scientific research. It should not be used for other purposes.')


import os
import json
from urllib.parse import quote
from typing import Union
from .mutable_local import set_mutable_local
from .store_file_local import _compute_file_hash
from .get_kachery_cloud_dir import get_kachery_cloud_dir


def link_file(filename: str, *, label: Union[str, None]=None):
    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)

    # this will be efficient on subsequent calls because of lookup
    sha1 = _compute_file_hash(filename, algorithm='sha1')

    size0 = os.path.getsize(filename)
    kachery_cloud_dir = get_kachery_cloud_dir()
    s = sha1
    parent_dir = f'{kachery_cloud_dir}/linked_files/sha1/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}'
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)
    with open(f'{parent_dir}/{sha1}', 'w') as f:
        f.write(json.dumps({
            'path': filename,
            'size': size0,
            'mtime': os.stat(filename).st_mtime,
            'sha1': sha1
        }))

    uri = f'sha1://{sha1}'
    if label is not None:
        uri = f'{uri}?label={quote(label)}'
    return uri


from ast import mod
import os
import yaml
from .get_kachery_cloud_dir import get_kachery_cloud_dir


_global = {
    'config': None
}

def _get_local_client_config():
    config = _global['config']
    if config is None:
        fname = f'{get_kachery_cloud_dir()}/config.yaml'
        config = {}
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                config = yaml.safe_load(f)
        modified = False
        # if 'multiuser' not in config:
        #     config['multiuser'] = False
        #     modified = True
        if modified:
            with open(fname, 'w') as f:
                yaml.safe_dump(config, f)
        _global['config'] = config
    return config


import requests
from .TemporaryDirectory import TemporaryDirectory


def _load_github_file(uri: str):
    from .store_file_local import store_file_local
    from .load_file import load_file_local
    
    user_name, repo_name, branch_name, file_name = _parse_github_uri(uri)
    url = f'https://raw.githubusercontent.com/{user_name}/{repo_name}/{branch_name}/{file_name}'
    with TemporaryDirectory(prefix='load_github_file') as tmpdir:
        tmp_filename = f'{tmpdir}/file.dat'
        with requests.get(url, stream=True) as r:
            if r.status_code == 404:
                raise Exception(f'File not found: {url}')
            r.raise_for_status()
            with open(tmp_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        uri = store_file_local(tmp_filename)
        return load_file_local(uri)

def _parse_github_uri(uri: str):
    if not uri.startswith('gh://'):
        raise Exception(f'Invalid github URI: {uri}')
    a = uri.split('/')
    if len(a) < 6:
        raise Exception(f'Invalid github URI: {uri}')
    user_name = a[2]
    repo_name = a[3]
    branch_name = a[4]
    file_name = '/'.join(a[5:])
    return user_name, repo_name, branch_name, file_name